# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Record, Draft and Parent Record API classes.

These classes belongs to the  data access layer and MUST ONLY be accessed from
within the service layer. It's wrong to use these classes in the presentation
layer.

A record and a draft share a single parent record. The parent record is used
to store properties common to all versions of a record (e.g. access control).

The draft and record share the same UUID, and thus both also share a single
persistent identifier. The parent record has its own UUID and own persistent
identifier.
"""

import uuid
from datetime import datetime

from invenio_pidstore.models import PIDStatus
from invenio_pidstore.providers.recordid_v2 import RecordIdProviderV2
from invenio_records.systemfields import ModelField
from invenio_records_resources.records import Record as RecordBase
from invenio_records_resources.records.systemfields import PIDField, PIDStatusCheckField
from sqlalchemy.orm.exc import NoResultFound

from .systemfields import ParentField, VersionsField


#
# Persistent identifier providers
#
class DraftRecordIdProviderV2(RecordIdProviderV2):
    """Draft PID provider."""

    default_status_with_obj = PIDStatus.NEW


#
# Record API classes
#
class ParentRecord(RecordBase):
    """Parent record API."""

    # Configuration
    model_cls = None

    pid = PIDField("id", provider=DraftRecordIdProviderV2, delete=True)


class Record(RecordBase):
    """Record API."""

    #: Class attribute to make it easy to check if record is a draft or not.
    is_draft = False

    #
    # Configuration to be set by a subclass
    #

    #: The record's SQLAlchemy model class. Must be set by the subclass.
    model_cls = None
    #: The parent state's SQLAlchemy model class. Must be set by the subclass.
    versions_model_cls = None
    #: The parent record's API class. Must be set by the subclass.
    parent_record_cls = None

    #
    # System fields
    #
    #: The internal persistent identifier. Records and drafts share UUID.
    pid = PIDField("id", provider=DraftRecordIdProviderV2, delete=True)

    #: System field to check if a record has been published.
    is_published = PIDStatusCheckField(status=PIDStatus.REGISTERED)

    #: The parent record - the draft is responsible for creating the parent.
    parent = ParentField(
        ParentRecord, create=False, soft_delete=False, hard_delete=False
    )

    #: Version relationship
    versions = VersionsField(create=True, set_latest=True)

    @classmethod
    def get_records_by_parent(cls, parent, include_deleted=True):
        """Get all sibling records for the specified parent record."""
        query = cls.model_cls.query.filter_by(parent_id=parent.id)
        if not include_deleted:
            query = query.filter_by(is_deleted=False)
        versions = query.all()

        return [
            cls(rec_model.data, model=rec_model, parent=parent)
            for rec_model in versions
        ]

    @classmethod
    def publish(cls, draft):
        """Publish a draft as a new record.

        If a record already exists, we simply get the record. If a draft has
        not yet been published, we create the record.

        The caller is responsible for registering the internal persistent
        identifiers (see ``register()``).
        """
        if draft.is_published:
            record = cls.get_record(draft.id)
        else:
            record = cls.create(
                {},
                # A draft and record share UUID, so we reuse the draft's id/pid
                id_=draft.id,
                pid=draft.pid,
                # Link the record with the parent record and set the versioning
                # relationship.
                parent=draft.parent,
                versions=draft.versions,
            )
            # Merge the PIDs into the current db session if not already in the
            # session.
            cls.pid.session_merge(record)
            cls.parent_record_cls.pid.session_merge(record.parent)
        return record

    def register(self):
        """Register the internal persistent identifiers."""
        if not self.parent.pid.is_registered():
            self.parent.pid.register()
            self.parent.commit()
        self.pid.register()


class Draft(Record):
    """Draft base API for metadata creation and manipulation."""

    #: Class attribute to make it easy to check if record is a draft or not.
    is_draft = True

    #
    # Configuration to be set by a subclass
    #

    #: The record's SQLAlchemy model class. Must be set by the subclass.
    model_cls = None
    #: The parent state's SQLAlchemy model class. Must be set by the subclass.
    versions_model_cls = None
    #: The parent record's API class. Must be set by the subclass.
    parent_record_cls = None

    #
    # System fields
    #

    #: The internal persistent identifier. Records and drafts share UUID.
    pid = PIDField("id", provider=DraftRecordIdProviderV2, delete=False)

    #: The parent record - the draft is responsible for creating the parent.
    parent = ParentField(ParentRecord, create=True, soft_delete=False, hard_delete=True)

    #: Version relationship
    versions = VersionsField(create=True, set_next=True)

    #: The expiry date of the draft.
    expires_at = ModelField()

    #: Revision id of record from which this draft was created.
    fork_version_id = ModelField()

    @classmethod
    def new_version(cls, record):
        """Create a draft for a new version of a record.

        The caller is responsible for:
        1) checking if a draft for a new version already exists
        2) moving the record data into the draft data.
        """
        return cls.create(
            {},
            # We create a new id, because this is for a new version.
            id=uuid.uuid4(),
            # Links the draft with the same parent (i.e. a new version).
            parent=record.parent,
            versions=record.versions,
            # New drafts without a record (i.e. unpublished drafts) must set
            # the fork version id to None.
            fork_version_id=None,
        )

    @classmethod
    def edit(cls, record):
        """Create a draft for editing an existing version of a record."""
        try:
            # We soft-delete a draft once it has been published, in order to
            # keep the version_id counter around for optimistic concurrency
            # control (both for search indexing and for REST API clients)
            draft = cls.get_record(record.id, with_deleted=True)
            if draft.is_deleted:
                draft.undelete()
                # Below line is needed to dump PID back into the draft.
                draft.pid = record.pid
                # Ensure record is link with the parent
                draft.parent = record.parent
                draft.versions = record.versions
                # Ensure we record the revision id we forked from
                draft.fork_version_id = record.revision_id
                # Note, other values like bucket_id values was kept in the
                # soft-deleted record, so we are not setting them again here.
        except NoResultFound:
            # If a draft was ever force deleted, then we will create the draft.
            # This is a very exceptional case as normally, when we edit a
            # record then the soft-deleted draft exists and we are in above
            # case.
            draft = cls.create(
                {},
                # A draft to edit a record must share the id and uuid.
                id_=record.id,
                pid=record.pid,
                # Link it with the same parent record
                parent=record.parent,
                versions=record.versions,
                # Record which record version we forked from.
                fork_version_id=record.revision_id,
            )
        return draft

    @classmethod
    def cleanup_drafts(cls, td):
        """Clean up (hard delete) all the soft deleted drafts.

        The drafts in the last timedelta span of time won't be deleted.
        """
        timestamp = datetime.utcnow() - td
        draft_model = cls.model_cls
        models = draft_model.query.filter(
            draft_model.is_deleted == True,  # noqa
            draft_model.updated < timestamp,
        ).all()

        # we need to clear the foreign keys in the version info
        for model in models:
            draft = cls(model.data, model=model)
            draft.versions.clear_next()

        # now we can delete the drafts without violating foreign keys
        ids = [model.id for model in models]
        draft_model.query.filter(draft_model.id.in_(ids)).delete(
            synchronize_session=False
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Parent record system field."""

from invenio_db import db
from invenio_records.systemfields import RelatedModelField
from sqlalchemy.exc import IntegrityError


class ParentField(RelatedModelField):
    """Parent record field."""

    def __init__(
        self, model, key=None, create=False, soft_delete=True, hard_delete=True
    ):
        """Initialize the parent field."""
        self._soft_delete = soft_delete
        self._hard_delete = hard_delete
        self._create = create
        super().__init__(
            model,
            key=key,
            required=True,
            dump=self.dump_parent,
            load=self.load_parent,
        )

    def create(self, record):
        """Method to create the parent record for the record."""
        parent = getattr(record, self.attr_name)
        if parent is None:
            # Create a parent record
            parent = record.parent_record_cls.create({})
            # This bumps revision id to 2 for the first record. Commit needed
            # because of the system fields defined on the parent record (e.g.
            # pid).
            parent.commit()

            # Set using the __set__() method
            setattr(record, self.attr_name, parent)
        return parent

    def delete(self, record, force=False):
        """Method to delete the parent record.

        The parent record will only be hard deleted if this is the last
        record/draft with links to it.
        """
        # A *record* is configured to never delete the parent
        # automatically.
        # A *draft* is configured to delete the parent on hard delete but not
        # on soft delete.
        #
        # A draft may be hard-deleted in two cases: 1) a new record 2) a new
        # record version. In the case 1, only a draft and the parent record
        # exists. In case 2, one or more drafts/records may exists with a
        # foreign key to the parent record.
        #
        # The logic implemented here, thus on a hard delete tries to delete the
        # the parent record and relies on the database integrity constraints
        # to prevent the parent record from being deleted in case more
        # drafts/records exists.
        parent = getattr(record, self.attr_name)
        if parent:
            if force and self._hard_delete:
                try:
                    with db.session.begin_nested():
                        parent.delete(force=True)
                except IntegrityError:
                    # It's ok - the draft/record linking to this parent is not
                    # the last one.
                    pass
            elif not force and self._soft_delete:
                parent.delete(force=False)

    #
    # Life-cycle hooks
    #
    def pre_create(self, record):
        """Called before a record is created (but after being initialized)."""
        if self._create:
            self.create(record)

    def post_delete(self, record, force=False):
        """Called after a record is deleted."""
        if self._hard_delete or self._soft_delete:
            self.delete(record, force=force)

    def pre_dump(self, record, data, dumper=None):
        """Called before a record is dumped."""
        parent = getattr(record, self.attr_name)
        # Use the default dumper for the parent (since the records/drafts
        # dumpers won't work on a parent)
        if parent:
            data[self.attr_name] = parent.dumps(dumper=None)

    def post_load(self, record, data, loader=None):
        """Called after a record was loaded."""
        parent_dump = record.pop(self.attr_name, None)
        if parent_dump:
            parent = record.parent_record_cls.loads(parent_dump, loader=loader)
            setattr(record, self.attr_name, parent)

    #
    # Helpers
    #
    @staticmethod
    def load_parent(field, record):
        """Serializer the object into a record."""
        if record.model.parent_id:
            return record.parent_record_cls.get_record(record.model.parent_id)
        return None

    @staticmethod
    def dump_parent(field, record, parent_record):
        """Set the object."""
        record.model.parent_id = parent_record.id



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Drafts-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Celery tasks to manage drafts."""

from datetime import timedelta

from celery import shared_task
from invenio_records_resources.proxies import current_service_registry


@shared_task(ignore_result=True)
def cleanup_drafts(seconds=3600):
    """Hard delete of soft deleted drafts.

    :param int seconds: numbers of seconds that should pass since the
        last update of the draft in order to be hard deleted.
    """
    timedelta_param = timedelta(seconds=seconds)
    service = current_service_registry.get("records")
    service.cleanup_drafts(timedelta_param)
